ns_log notice "Loading ns_xml-rpc.tcl"

# ns_xml-rpc 2001/03/30
# xml-rpc server and client implementation for AOLserver/OpenNSD
# with ns_xml module
# Dave Bauer dave@thedesignexperience.org
# taken from xmlrpc.tcl from Steve Ball

#
# setup nsv array to hold procs that are registered for xmlrpc

nsv_array set xmlrpc_procs [list]


# xmlrpc_register_proc
# 
# register a proc to be available via XML-RPC
# 
# for now you have to call xmlrpc_register_proc for every
# proc that you want to be available
# 

proc xmlrpc_register_proc {proc_name} {
    nsv_set xmlrpc_procs $proc_name $proc_name
}

# xmlrpc_respond
#
# format a sucess response to a XML-RPC request
#
# Arguments
#   data   data to be returned to the client
#
# Results
#   return appropriately formatted text
#

proc xmlrpc_respond {data} {
    set response  [ns_xml doc create "1.0"]
    set doc_id    [ns_xml doc new_root $response methodResponse ""]
    set params_id [ns_xml node new_child $doc_id params ""]
    set param_id  [ns_xml node new_child $params_id param ""]
    set value_id  [ns_xml node new_child $param_id value ""]
    
    xmlrpc_construct $value_id {} $data

    set result [ns_xml doc render $response]
    ns_xml doc free $response

    return $result
}


# xmlrpc_fault
#   format a fault response to a XML-RPC request
#
# Arguments
#   code  error code
#   msg   error message
#
# Results
#   returns appropriately formatted text
#

proc xmlrpc_fault {code msg} {

    set response [ns_xml doc create "1.0"]
    set value \
            [ns_xml node new_child \
                [ns_xml node new_child \
                [ns_xml doc new_root $response methodResponse ""] \
                fault ""] value ""]

    set struct [ns_xml node new_child $value struct ""]
    set codemember [ns_xml node new_child $struct member ""]
    ns_xml node new_child \
            [ns_xml node new_child $codemember value ""] \
            int $code
    set msgmember [ns_xml node new_child $struct member ""]
    ns_xml node new_child \
            [ns_xml node new_child $msgmember value ""] \
            string $msg

    set result [ns_xml doc render $response]

    ns_xml doc free $response

    return $result
}

# xmlrpc_invoke
#
# the methodName element contains the Tcl procedure to evaluate
# the parameters are passed as arguments to the global eval
#
# Arguments
#   xml     XML-RPC data from the client
#   
# Results
#   return value of the method is encoded ready for return to
#   the client
#

proc xmlrpc_invoke {xml} {

    if {[catch {ns_xml parse $xml} doc_id]} {
        set result [xmlrpc_fault 1 "error parsing request: $doc_id"]
        ns_log error "xmlrpc_invoke: error parsing request: $doc_id"
    } else {
        set data   [ns_xml doc root $doc_id]
	if {[ns_xml node name $data] != "methodCall"} {
	    set data [lindex [ns_xml node children $data] 0]
	}
        set methodName \
            [ns_xml node getcontent \
              [lindex \
                [ns_xml_getElementsByTagName $data methodName] 0 ]]

# check that the method is registered as a  valid XML-RPC method 
#
	
	if {![nsv_exists xmlrpc_procs $methodName]} {
	     set result [xmlrpc_fault "00" "methodName does not exist"]
	     return $result
	    }

        set arguments [list]
        set params [ns_xml_getElementsByTagName $data params]
        foreach parameter \
                [ns_xml_getElementsByTagName $params param] {

            lappend arguments \
              [xmlrpc_decodeValue \
                 [lindex [ns_xml_getChildrenTrim $parameter] 0]]
        }
    
        if {[catch {uplevel #0 [list $methodName] $arguments} result]} {
        # don't ask me what the $:: means
            set result [xmlrpc_fault "1" $result]
            ns_log error "xmlrpc_invoke: error in xmlrpc method $methodName $result"
        } else {
            set result [xmlrpc_respond $result]
        }
    }
    return $result
}



# xmlrpc_createContext
#
#   create a container element for data
#
#   Arguments
#
#     parent  parent node
#
#     context context to create 
#
# 
proc xmlrpc_createContext { parent context } {
    
    foreach child $context {
        set parent \
                [ns_xml node new_child $parent $child ""]
    }

    return $parent
}

# xmlrpc_callResponse
#
# Arguments
#   request HTTP request token
#
# Result
#   returns the result of the XML-RPC request

proc xmlrpc_callResponse {response} {
    # ns_log debug xmlrpc_CallResponse response \"$response\"
    if {[catch {xmlrpc_parse $response} result]} {
        return -code error \
                "XMLRPC-ERROR remote procedure call failed: \"$result\""
    }
  
    return $result
}

# xmlrpc_call
#
# invoke a method on the remote server using XML-RPC
#
# Arguments
#   url     url of service
#   method  method to call
#   args    args to the method
#
# Results
#   return the response of the service or an error if
#   the service returns a fault

proc xmlrpc_call {url method args} {
    set rpc [ns_xml doc create "1.0"]
    set cleanup [list ns_xml doc free $rpc]
    set call [ns_xml doc new_root $rpc methodCall ""]
    ns_xml node new_child $call methodName $method
    
    set params [ns_xml node new_child $call params ""]

    if {[catch {
        xmlrpc_construct $params {param value} $args
    } errMsg]} {
        eval $cleanup
        return -code error $errMsg
    }

    #make the call
    # experimental, how can we use util_httppost 
    
    if {[catch {
        set request [ns_xml doc render $rpc]
        # ns_log debug xmlrpc_call url $url request \"$request\"
        util_httppost $url $request
    } response ]} {
        eval $cleanup
        ns_log error xmlrpc_call url $url methodName $method error: $response
        return -code error \
                [list HTTP_ERROR \
                "HTTP request failed due to \"$response\""]
    }

    eval $cleanup
    # skip over httpoptions part...
    # that is for the HTTP package for Tcl

    return [xmlrpc_callResponse $response]
}



# xmlrpc_construct --
#
#     Construct the XML-RPC resquest or response
#
# Arguments
#     node node to add child elements to
#     context     how to add the data
#     arglist     configuration options?
#
# Results
#     Nodes added to document tree. Returns unrecognized options
#

proc xmlrpc_construct { node context arglist } {
    set unused {}
    # consume string arguments until configuration options found 
    set options_list [list "-string" "-text" "-i4" "-int" "-integer" \
			  "-boolean" "-double" "-date" "-binary" "-base64" \
			  "-variable" "-structvariable" "-struct" \
			  "-array" "-keyvalue"]
    # we set the list of valid options, some tcl users
    # choose to pass strings containing a leading -
    
    while {[llength $arglist] && \
            [lsearch $options_list [lindex $arglist 0]]!=-1} {

        ns_xml node setcontent \
                [xmlrpc_createContext $node $context] \
                [lindex $arglist 0]
        set arglist [lreplace $arglist 0 0]
    }

    # Now process configuration options
    
    if { [llength $arglist] % 2} {
        return -code error \
                "no value for option \"[lindex $arglist end]\""
    }
    
    foreach {option value} $arglist {
        switch -- $option {
            -string -
            -text {
                ns_xml node setcontent \
                        [xmlrpc_createContext $node $context] \
                        $value
            }

            -i4 -
            -int -
            -integer {
                if {![string is integer $value]} {
                    return -code error \
                            "value \"$value\" for option \"$option\" is not an integer:"
                }
                ns_xml node new_child \
                        [xmlrpc_createContext $node $context] \
                        i4 $value
            }

            -boolean {
                ns_xml node new_child \
                        [xmlrpc_createContext $node $context] \
                        boolean [string is true $value]
            }

            -double {
                if {![string is double $value]} {
                    return -code error \
                            "value \"$value\" for option \"$option\" is not a floating point value"
                }
                ns_xml node new_child \
                        [xmlrpc_createContext $node $context] \
                        double $value
            }

            -date {

                if {[catch {
                    clock format [clock scan $value] \
                          -format {%Y%m%dT%H:%M:%S}
                } datevalue]} {
                    return -code error \
                            "value \"$value\" for option \"$option\" is not a valid date ($datevalue)"
                }
                
                ns_xml node new_child \
                        [xmlrpc_createContext $node $context] \
                        "dateTime.iso8601" $datevalue
            }

            -binary -
            -base64 {
                
                # it is up to the application to do the
                # encoding, I think it's built into AOLserver.. 
                # anyway it should be done before the 
                # data is sent here

                ns_xml node new_child \
                        [xmlrpc_createContext $node $context] \
                        base64 $value

            }

            -variable {
                
                upvar 2 $value var
                if {[array exists var]} {
                    set data nx_xml node new_child \
                        [ns_xml node new_child \
                            [xmlrpc_createContext $node $context] \
                                array "" ] data "" ]

                    foreach {idx entry [array get var]} {
                        nx_xml node new_child $data value $entry
                    }
                } else {
                    ns_xml node setcontent \
                        [xmlrpc_createContext $node $context] \
                        $value 
            }   
        }

        -structvariable {
            
            upvar 2 $value var
            set struct [ns_xml node new_child \
                [xmlrpc_createContext $node $context] $struct ""]
           
            foreach [idx entry] [array get var] {
                set member \
                        [ns_xml node newchild $struct member ""]
                ns_xml node newchild $member name $idx
                ns_xml node newchild $member value $entry
            }
        }

        -array {

            set data \
              [ns_xml node new_child \
                [ns_xml node new_child \
                   [xmlrpc_createContext $node $context] \
                   array ""] \
                data ""]
            foreach datum $value {
                set result [xmlrpc_construct $data value $datum]
                if {[llength $result]} {
                    return -code error \
                            "unknown configuration option \"[lindex $result 0]\""
                }
            }
        }

        -struct -
        -keyvalue {
            set struct \
                    [ns_xml node new_child \
                    [xmlrpc_createContext $node $context] \
                    struct ""]
            foreach {name mvalue} $value {
                set member \
                    [ns_xml node new_child $struct member ""]
                ns_xml node new_child $member name $name
                set result \
                        [xmlrpc_construct $member value $mvalue]
                if {[llength $result]} {
                    return -code error \
                            "unkown configuration option \"[lindex $result 0]\""
                }
            }
        }

        default {
            #anything else will be passed back
            lappend unused $option $value
        }
    }
}

return $unused
}


# xmlrpc_decodeValue
#   Unpack the data in a value element
#
# Arguments 
#   node   value element node
#
# Results
#   Returns data. If the value is a struct then returns the data
#   in name-value pairs. If the value is an array then returns
#   the data as name-value pairs where the name is an integer
#   starting with 0.

proc xmlrpc_decodeValue {node} {

    set result ""

    if {[llength [ns_xml node children $node]] != 0} {  

        set nodeType [ns_xml node type [ns_xml_firstChild $node]]
        # ns_log debug decodeValue nodeType $nodeType
        switch -- $nodeType {
            cdata_section {
                set result \
                        [ns_xml node getcontent \
                        [ns_xml_firstChild  $node]]
                # ns_log debug cdata result $result
            }
            
            attribute {
                set attrType [ns_xml node name \
                        [ns_xml_firstChild $node]] 
                switch -- $attrType {
                    string -
                    i4 -
                    int -
                    double -
                    boolean  {
                        
                        set result \
                                [ns_xml node getcontent \
                                [ns_xml_firstChild $node]]
                    }
                    
                    dateTime.iso8601 {
                        set result \
                                [clock scan \
                                [ns_xml node getcontent \
                                [ns_xml_firstChild $node]]]
                    }
                    
                    base64 {
                        set result \
                                [ns_xml node getcontent \
                                [ns_xml_firstChild $node]]
                    }
                    
                    struct {
                        foreach member \
                                [ns_xml_getElementsByTagName \
                                [ns_xml_firstChild $node] member] {
                            lappend result \
                                    [ns_xml node getcontent \
                                    [ns_xml_getElementsByTagName \
                                    $member name]]
                            set stuff \
                                    [xmlrpc_decodeValue \
                                    [ns_xml_getElementsByTagName \
                                    $member value]]
#                              ns_log notice "XMLRPC:stuff:$stuff"
			    lappend result $stuff
			}
		    }

		    array {
			set index 0

			foreach entry \
				[ns_xml_getChildrenTrim \
				[ns_xml_getElementsByTagName \
				[ns_xml_firstChild $node] data ]]  {
			    lappend result [incr index]
			    lappend result \
				    [xmlrpc_decodeValue $entry]
			}
		    }
		    
		}
	    }
	    
	    default {
		ns_log notice "XMLRPC:decode:node type not found"
	    }
	    
	}
    }
    return $result
}

proc xmlrpc_parse {xml} {

    set doc_id [ns_xml parse $xml]
    set response [ns_xml doc root $doc_id]

    if {[ns_xml node name $response] != "methodResponse"} {
	set response [lindex [ns_xml node children $response] 0]
    }
    
    set top [ns_xml_firstChild $response]
    switch -- [ns_xml node name $top] {
        params {
            set param [ns_xml_firstChild $top]
            #the above should be checked: node type, tag names, etc...
            set firstChild [ns_xml_firstChild $param]
            set result [xmlrpc_decodeValue $firstChild]
        }
        fault {
            # should do more checking here...
            # does value/struct/member, etc, exist? and so on
            array set fault [xmlrpc_decodeValue [ns_xml_firstChild $top]]
            return -code error [list $fault(faultCode) $fault(faultString)]
        }
        default {
            set type [ns_xml node name $response]
            return -code error "invalid server reposnse ($type)"
        }
    }
    return $result
}

ns_log notice "Done ns_xml-rpc.tcl"
