# XML-RPC Validation Test
# Written by, Aaron Swartz <aaron@theinfo.org>
# Enhanced by Jerry Asher <jerry@theashergroup.com>
# contains:
#   xml-rpc query dispatcher (registered proc)
#   xml-rpc client validator test implementation
#   xml-rpc server validation implementation

############################################################

# For some reason, AOLserver doesn't have an [ns_conn content]
# function.  It looks a bit like it was taken out at the last
# minute.  I have to talk about the AOLserver people about this.
# In the meantime, I have this hack that does the same:

proc getContent {} {

    # (taken from aol30/modules/tcl/form.tcl)
    # Spool content into a temporary read/write file.
    # ns_openexcl can fail, since tmpnam is known not to
    # be thread/process safe.  Hence spin till success
    set fp ""
    while {$fp == ""} {
        set filename "[ns_tmpnam][clock clicks].xmlrpc2"
        set fp [ns_openexcl $filename]
    }

    fconfigure $fp -translation binary
    ns_conncptofp $fp
    close $fp

    set fp [open $filename r]
    while {![eof $fp]} {
        append text [read $fp]
    }
    close $fp
    ns_unlink $filename
    return $text
}

proc xml_rpcdispatcher {} {
    if {[ns_conn method] == "GET"} {
        ns_return 200 text/html "XML-RPC Validator App
        <p>This is an
        <a href=\"http://xml-rpc.com\">XML-RPC</a> application.
        If you want to use it, try a POST request.
        </p>
        "
    } else {
        global content
        set content [getContent]
        ns_return 200 text/xml [xmlrpc_invoke $content]
    }
}

ns_register_proc GET  /RPC2 xml_rpcdispatcher
ns_register_proc POST /RPC2 xml_rpcdispatcher

############################################################
# The xml-rpc client validator procedures begin here:

# Takes an array, each of whose members is a struct.  Return the
# sum of all the values named curly from each struct.


proc validator1.arrayOfStructsTest {params} {
        
    set number 0
    foreach {ignore param} $params {
        array set struct $param
        set number [expr $number + $struct(curly)]
    }
    return [list -int $number]

    # Turn the list-ified array into a normal array (bigArray)
    array set bigArray [lindex $args 0]
    
    set counter 1
    set number 0
    # Loop
    while {[info exists bigArray($counter)]} {
        # De-list-ify the struct:
        array set struct $bigArray($counter)
        # Add curly to the number count:
        set number [expr $number + $struct(curly)]
        # Increment the counter before going on:
        incr counter
    }
    # Return the number count as an integer:
    return [list -int $number]
}

xmlrpc_register_proc validator1.arrayOfStructsTest

# Takes a string.
# Return the number of each entity in a struct.

proc validator1.countTheEntities {args} {
    # HACK: The XML-RPC parse doesn't deal with entities
    # very well. Instead, we get things directly.

    global content
    regexp {<value>(.*)</value>} $content {} string
    set string [util_expand_entities $string]
        
    # NORMAL VERSION: 
    #set string $args
        
    # For each type of entity, do a regsub -all and return the result as an integer,
    # then place it all in a struct with the proper names and return it.
    return \
      [list -struct \
        [list \
          ctLeftAngleBrackets  [list -int [regsub -all {\<} $string "" string]] \
          ctRightAngleBrackets [list -int [regsub -all {\>} $string "" string]] \
          ctAmpersands         [list -int [regsub -all {&}  $string "" string]] \
          ctApostrophes        [list -int [regsub -all {\'} $string "" string]] \
          ctQuotes             [list -int [regsub -all {\"} $string "" string]]]]
}

xmlrpc_register_proc validator1.countTheEntities

# Takes a struct.
# Return the sum of the values larry, curly and moe.

proc validator1.easyStructTest {struct} {
    # De-list-ify the stuct:
    array set bigStruct $struct
    # Return the sum as an integer:
    return [list -int [expr \
            $bigStruct(moe) \
            + $bigStruct(curly) \
            + $bigStruct(larry)]]
}

xmlrpc_register_proc validator1.easyStructTest

proc validator1.echoStructTest {struct} {
    foreach {name value} $struct {
        if {[llength $value] > 1} {
            # For the substructs:
            foreach {name2 value2} $value {
                set returnArray($name2) [list -int $value2]
            }
            set output($name) [list -struct [array get returnArray]]
        } else {
            set output($name) $value
        }
        array set returnArray ""
    }
    return [list -struct [array get output]]
}

xmlrpc_register_proc validator1.echoStructTest

proc validator1.manyTypesTest {
    number boolean string double dateTime base64
} {
    return [list -array \
             [list \
               [list -int $number] \
               [list -boolean $boolean] \
               [list -text $string] \
               [list -double $double] \
               [list -date [clock format $dateTime]] \
               [list -base64 $base64]]]
}

xmlrpc_register_proc validator1.manyTypesTest

proc validator1.moderateSizeArrayCheck {array} {
    array set bigArray $array
    set counter 1
    while {[info exists bigArray($counter)]} {
        incr counter
    }
    set counter [expr $counter - 1]
    return "-string [list "$bigArray(1)$bigArray($counter)"]"
}

xmlrpc_register_proc validator1.moderateSizeArrayCheck

proc validator1.nestedStructTest {struct} {
    array set bigStruct $struct
    array set 2000 $bigStruct(2000)
    array set April $2000(04)
    array set first $April(01)
    return "-int [expr $first(larry) \
            + $first(curly) + $first(moe)]"
}

xmlrpc_register_proc validator1.nestedStructTest

proc validator1.simpleStructReturnTest {number} {
    set struct(times10) [expr $number * 10]
    set struct(times100) [expr $number * 100]
    set struct(times1000) [expr $number * 1000]

    return "-struct [list [array get struct]]"
}

xmlrpc_register_proc validator1.simpleStructReturnTest

############################################################
# XML-RPC Server Validator
# change URL to the server you are trying to validate! for each of
# these procs..

proc validate1.arrayOfStructsTest {
{url http://www.theashergroup.com/RPC2}
    {array ""}
} {
    if {[string equal "" $array]} {
        set array [list \
              [list -struct [list moe [list -int 1] \
                                  curly [list -int 2] \
                                  larry [list -int 3]]] \
              [list -struct [list moe [list -int 1] \
                                  curly [list -int 2] \
                                  larry [list -int 3]]] \
              [list -struct [list moe [list -int 1] \
                                  curly [list -int 2] \
                                  larry [list -int 3]]]]
    }
    return [xmlrpc_call $url validator1.arrayOfStructsTest -array $array]
}

proc validate1.easyStructTest {
    {url "http://www.theashergroup.com/RPC2"}
    {struct ""}
} {
    if {[string equal "" $struct]} {
        set struct \
              [list moe [list -int 1] \
                    curly [list -int 2] \
                    larry [list -int 3]]

    }
    return [xmlrpc_call $url validator1.easyStructTest -struct $struct]
}

proc validate1.echoStructTest {
    {url "http://www.theashergroup.com/RPC2"}
    {struct ""}
} {
    if {[string equal $struct ""]} {
        set struct [list bob [list -int 5]]
    }
    return [xmlrpc_call $url validator1.echoStructTest -struct $struct]
}

proc validate1.manyTypesTest {
    {url http://www.theashergroup.com/RPC2}
    {int 1} 
    {boolean 0}
    {string wazzup}
    {double 3.14159}
    {date "20010704T11:50:30"}
    {base64 "R0lGODlhFgASAJEAAP/////OnM7O/wAAACH5BAEAAAAALAAAAAAWABIAAAJAhI+py40zDIzujEDBzW0n74AaFGChqZUYylyYq7ILXJJ1BU95l6r23RrRYhyL5jiJAT/Ink8WTPoqHx31im0UAAA7"}
} {
    return [xmlrpc_call $url validator1.manyTypesTest \
              -int $int -boolean $boolean -string $string \
              -double $double -date $date -base64 $base64]
}

proc validate1.nestedStructTest {
    {url http://www.theashergroup.com/RPC2}
    {moe 1}
    {larry 2}
    {curly 4}
    {startyear 1999}
    {endyear 2001}
} {
    
    set calendar ""
    # for each year
    for {set y $startyear} {$y <= $endyear} {incr y} {

        set year [list]
        # for each month
        for {set m 1} {$m <= 12} {incr m} {

            set month [list]
            # for each day
            set mstr [format %02d $m]
            for {set d 1} {$d <= 31} {incr d} {
                set dstr [format %02d $d]
                # exit test (to find end of month)
                set date \
                        [clock format \
                          [clock scan "[expr $d - 1] day" \
                            -base [clock scan "$y-${mstr}-01"]] \
                          -format "%y:%m:%d"]
                set date [split $date :]
                set reald [lindex $date 2]
                if {![string equal $reald $dstr]} {
                    break
                }
                
                if {($y == 2000) && ($m == 4) && ($d == 1)} {
                    set dayta \
                        [list -struct \
                           [list moe [list -int $moe] \
                                 curly [list -int $curly] \
                                 larry [list -int $larry]]]
                } else {
                    set dayta \
                        [list -struct \
                           [list moe [list -int [expr 2 * $moe]]]]
                }
                set month [concat $month [list $dstr $dayta]]
            }
            set year [concat $year       [list $mstr [list -struct $month]]]
        }
        set calendar [concat $calendar   [list $y    [list -struct $year]]]
    }
    
    return [xmlrpc_call $url validator1.nestedStructTest -struct $calendar]
}


proc validate1.simpleStructReturnTest {
    {url http://www.theashergroup.com/RPC2}
    {number 2}
} {
    return [xmlrpc_call $url validator1.simpleStructReturnTest -int $number]
}

